# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sender_key.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10sender_key.proto\x12\ntextsecure\"\xb8\x03\n\x17SenderKeyStateStructure\x12\x13\n\x0bsenderKeyId\x18\x01 \x01(\r\x12J\n\x0esenderChainKey\x18\x02 \x01(\x0b\x32\x32.textsecure.SenderKeyStateStructure.SenderChainKey\x12N\n\x10senderSigningKey\x18\x03 \x01(\x0b\x32\x34.textsecure.SenderKeyStateStructure.SenderSigningKey\x12O\n\x11senderMessageKeys\x18\x04 \x03(\x0b\x32\x34.textsecure.SenderKeyStateStructure.SenderMessageKey\x1a\x31\n\x0eSenderChainKey\x12\x11\n\titeration\x18\x01 \x01(\r\x12\x0c\n\x04seed\x18\x02 \x01(\x0c\x1a\x33\n\x10SenderMessageKey\x12\x11\n\titeration\x18\x01 \x01(\r\x12\x0c\n\x04seed\x18\x02 \x01(\x0c\x1a\x33\n\x10SenderSigningKey\x12\x0e\n\x06public\x18\x01 \x01(\x0c\x12\x0f\n\x07private\x18\x02 \x01(\x0c\"X\n\x18SenderKeyRecordStructure\x12<\n\x0fsenderKeyStates\x18\x01 \x03(\x0b\x32#.textsecure.SenderKeyStateStructure')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sender_key_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SENDERKEYSTATESTRUCTURE._serialized_start=33
  _SENDERKEYSTATESTRUCTURE._serialized_end=473
  _SENDERKEYSTATESTRUCTURE_SENDERCHAINKEY._serialized_start=318
  _SENDERKEYSTATESTRUCTURE_SENDERCHAINKEY._serialized_end=367
  _SENDERKEYSTATESTRUCTURE_SENDERMESSAGEKEY._serialized_start=369
  _SENDERKEYSTATESTRUCTURE_SENDERMESSAGEKEY._serialized_end=420
  _SENDERKEYSTATESTRUCTURE_SENDERSIGNINGKEY._serialized_start=422
  _SENDERKEYSTATESTRUCTURE_SENDERSIGNINGKEY._serialized_end=473
  _SENDERKEYRECORDSTRUCTURE._serialized_start=475
  _SENDERKEYRECORDSTRUCTURE._serialized_end=563
# @@protoc_insertion_point(module_scope)
